@@ -17,11 +17,14 @@
 #include <grp.h>
 #include <sys/stat.h>
 #include <syslog.h>
+#include <libgen.h>
 
 #include "config.h"
 #include "miniobj.h"
 #include "configuration.h"
 #include "vas.h"
+#include "vsb.h"
+#include "cfg_parser.h"
 
 #define ADDR_LEN 150
 #define PORT_LEN 6
@@ -76,6 +79,9 @@
 
 #define CFG_CONFIG "config"
 
+extern FILE *yyin;
+extern int yyparse(hitch_config *);
+
 // END: configuration parameters
 
 static char error_buf[CONFIG_BUF_SIZE];
@@ -192,13 +198,12 @@ config_destroy(hitch_config *cfg)
 		free(fa->ip);
 		free(fa->port);
 		free(fa->pspec);
-		CHECK_OBJ_ORNULL(fa->cert, CFG_CERT_FILE_MAGIC);
-		if (fa->cert != NULL) {
-			fa->cert->ref--;
-			if (fa->cert->ref == 0) {
-				free(fa->cert->filename);
-				FREE_OBJ(fa->cert);
-			}
+		free(fa->ciphers);
+		HASH_ITER(hh, fa->certs, cf, cftmp) {
+			CHECK_OBJ_NOTNULL(cf, CFG_CERT_FILE_MAGIC);
+			HASH_DEL(fa->certs, cf);
+			free(cf->filename);
+			FREE_OBJ(cf);
 		}
 		FREE_OBJ(fa);
 	}
@@ -234,47 +239,6 @@ config_destroy(hitch_config *cfg)
 	free(cfg);
 }
 
-int
-config_parse_content(char *line, char **key, char **value)
-{
-	assert(line != NULL);
-
-	if (line[0] == '#')
-		return 1;  // NOOP
-
-	if (strlen(line) < 1 || line[0] == '\n' || strcmp(line, "\r\n") == 0)
-		return 1;
-
-	while (*line != '\0' && isspace(*line)) line++;
-	*key = line;
-	while(*line != '\0' && (isalnum(*line) || *line == '-')) line++;
-	if (*line == '\0' || *(line+1) == '\0')
-		return -1;
-	*line = '\0'; // key end.
-	line++;
-
-	while(*line != '\0' && (*line != '=')) line++;
-	if (*line != '=')
-		return -3;
-
-	if (*line == '\0' || *(line+1) == '\0')
-		return -1;
-	line++;
-
-	while(*line != '\0' && (isspace(*line) || *line == '"' || *line == '\'')) line++;
-	if (*line == '\0')
-		return -1;
-	*value = line;
-
-	while (*line != '\0' && *line != '"' && *line != '\'' && !isspace(*line)) line++;
-	*line = '\0';  // value end.
-
-	if (strlen(*key) <= 1 || strlen(*value) < 1)
-		return -1;
-
-	return(0);
-}
-
 char *
 config_assign_str(char **dst, char *v)
 {
@@ -459,11 +423,15 @@ config_param_pem_file(char *filename, struct cfg_cert_file **cfptr)
 	ALLOC_OBJ(cert, CFG_CERT_FILE_MAGIC);
 	AN(cert);
 	config_assign_str(&cert->filename, filename);
-	cert->mtim = st.st_mtim.tv_sec
-	    + st.st_mtim.tv_nsec * 1e-9;
+	cert->mtim = st.st_mtime;
+#if defined(HAVE_STRUCT_STAT_ST_MTIM)
+	cert->mtim += st.st_mtim.tv_nsec * 1e-9;
+#elif defined(HAVE_STRUCT_STAT_ST_MTIMESPEC)
+	cert->mtim += st.st_mtimespec.tv_nsec * 1e-9;
+#endif
+
 
 	*cfptr = cert;
-	cert->ref++;
 	return (1);
 
 }
@@ -573,6 +541,67 @@ config_param_shcupd_peer(char *str, hitch_config *cfg)
 
 #endif /* USE_SHARED_CACHE */
 
+struct front_arg *
+front_arg_new(void)
+{
+	struct front_arg *fa;
+
+	ALLOC_OBJ(fa, FRONT_ARG_MAGIC);
+	AN(fa);
+	fa->match_global_certs = -1;
+	fa->sni_nomatch_abort = -1;
+	fa->etype = ENC_TLS;
+	fa->prefer_server_ciphers = -1;
+
+	return (fa);
+}
+
+int
+front_arg_add(hitch_config *cfg, struct front_arg *fa)
+{
+	struct vsb pspec;
+
+	CHECK_OBJ_NOTNULL(fa, FRONT_ARG_MAGIC);
+	AN(fa->port);
+
+	if (cfg->LISTEN_DEFAULT != NULL) {
+		/* drop default listen arg. */
+		struct front_arg *def = NULL;
+		HASH_FIND_STR(cfg->LISTEN_ARGS, "default", def);
+		AN(def);
+		HASH_DEL(cfg->LISTEN_ARGS, def);
+		free(def->ip);
+		free(def->port);
+		free(def->pspec);
+		FREE_OBJ(def);
+		cfg->LISTEN_DEFAULT = NULL;
+	}
+
+	VSB_new(&pspec, NULL, 0, VSB_AUTOEXTEND);
+	VSB_printf(&pspec, "[%s]:", fa->ip);
+	VSB_cat(&pspec, fa->port);
+	VSB_finish(&pspec);
+	fa->pspec = VSB_data(&pspec);
+	HASH_ADD_KEYPTR(hh, cfg->LISTEN_ARGS, fa->pspec,
+	    strlen(fa->pspec), fa);
+
+	if (fa->match_global_certs == -1) {
+		if (HASH_CNT(hh, fa->certs) == 0)
+			fa->match_global_certs = 1;
+		else
+			fa->match_global_certs = 0;
+	} else {
+		if (HASH_CNT(hh, fa->certs) == 0
+		    && fa->match_global_certs == 0) {
+			config_error_set("No certificate configured "
+			    "for frontend '%s'", fa->pspec);
+			return (0);
+		}
+	}
+
+	return (1);
+}
+
 int
 config_param_validate(char *k, char *v, hitch_config *cfg,
     char *file, int line)
@@ -603,36 +632,21 @@ config_param_validate(char *k, char *v, hitch_config *cfg,
 		struct cfg_cert_file *cert;
 		char *certfile = NULL;
 
-		ALLOC_OBJ(fa, FRONT_ARG_MAGIC);
+		fa = front_arg_new();
 		r = config_param_host_port_wildcard(v,
 		    &fa->ip, &fa->port, &certfile, 1);
 		if (r != 0) {
-			if (cfg->LISTEN_DEFAULT != NULL) {
-				/* drop default listen arg. */
-				struct front_arg *def = NULL;
-				HASH_FIND_STR(cfg->LISTEN_ARGS, "default", def);
-				AN(def);
-				HASH_DEL(cfg->LISTEN_ARGS, def);
-				free(def->ip);
-				free(def->port);
-				free(def->cert);
-				free(def->pspec);
-				FREE_OBJ(def);
-				cfg->LISTEN_DEFAULT = NULL;
-			}
-			fa->pspec = strdup(v);
-			HASH_ADD_KEYPTR(hh, cfg->LISTEN_ARGS, fa->pspec,
-			    strlen(fa->pspec), fa);
 			if (certfile != NULL) {
 				r = config_param_pem_file(certfile, &cert);
 				if (r != 0) {
 					AN(cert);
-					fa->cert = cert;
+					HASH_ADD_KEYPTR(hh,
+					    fa->certs, cert->filename,
+					    strlen(cert->filename), cert);
 				}
 				free(certfile);
-			}
-		} else {
-			FREE_OBJ(fa);
+		}
+			r = front_arg_add(cfg, fa);
 		}
 	} else if (strcmp(k, CFG_BACKEND) == 0) {
 		free(cfg->BACK_PORT);
@@ -789,45 +803,33 @@ config_param_validate(char *k, char *v, hitch_config *cfg,
 int
 config_file_parse(char *file, hitch_config *cfg)
 {
-	char line[CONFIG_BUF_SIZE];
-	char *key, *value;
-	FILE *fd = NULL;
-
-	int r;
+	FILE *fp = NULL;
+	int r = 0;
 
 	AN(cfg);
 
 	// should we read stdin?
 	if (file == NULL || strlen(file) < 1 || strcmp(file, "-") == 0)
-		fd = stdin;
+		fp = stdin;
 	else
-		fd = fopen(file, "r");
+		fp = fopen(file, "r");
 
-	if (fd == NULL) {
+	if (fp == NULL) {
 		config_error_set("Unable to open configuration file '%s': %s\n",
 		    file, strerror(errno));
 		return (1);
 	}
 
-	int i = 0;
-	while (1) {
-		if (fgets(line, sizeof(line)-1, fd) == NULL)
+	yyin = fp;
+	do {
+		if (yyparse(cfg) != 0) {
+			r = 1;
 			break;
-		i++;
-
-		r = config_parse_content((char*)&line, &key, &value);
-		if (r != 0) 	/* comments/blank lines */
-			continue;
-		// printf("File '%s', line %d, key: '%s', value: '%s'\n", file, i, key, value);
-
-		if (config_param_validate(key, value, cfg, file, i) != 0) {
-			fclose(fd);
-			return (1);
 		}
-	}
+	} while (!feof(yyin));
 
-	fclose(fd);
-	return (0);
+	fclose(fp);
+	return (r);
 }
 
 char *
@@ -1109,7 +1111,7 @@ config_parse_cli(int argc, char **argv, hitch_config *cfg, int *retval)
 			ret = config_param_validate(CFG_SHARED_CACHE_LISTEN, optarg, cfg, NULL, 0);
 			break;
 		case 'P':
-			ret = config_param_validate(CFG_SHARED_CACHE_PEER, optarg, cfg, NULL, 0)s;
+			ret = config_param_validate(CFG_SHARED_CACHE_PEER, optarg, cfg, NULL, 0);
 			break;
 		case 'M':
 			ret =config_param_validate(CFG_SHARED_CACHE_MCASTIF, optarg, cfg, NULL, 0);
@@ -1212,10 +1214,19 @@ config_parse_cli(int argc, char **argv, hitch_config *cfg, int *retval)
 			return (1);
 		}
 	}
+
+
 	if (cfg->PMODE == SSL_SERVER && cfg->CERT_DEFAULT == NULL) {
-		config_error_set("No x509 certificate PEM file specified!");
-		*retval = 1;
-		return (1);
+		struct front_arg *fa, *fatmp;
+		HASH_ITER(hh, cfg->LISTEN_ARGS, fa, fatmp) {
+			if (HASH_CNT(hh, fa->certs) == 0) {
+				config_error_set("No x509 certificate PEM file "
+				    "specified for frontend '%s'!", fa->pspec);
+				*retval = 1;
+				return (1);
+
+			}
+		}
 	}
 
 	return (0);
