@@ -152,24 +152,35 @@ struct child_proc {
 VTAILQ_HEAD(child_proc_head, child_proc);
 static struct child_proc_head child_procs;
 struct sslctx_s;
+struct sni_name_s;
 
 struct listen_sock {
-	unsigned			magic;
-#define LISTEN_SOCK_MAGIC 		0x5b04e577
+	unsigned		magic;
+#define LISTEN_SOCK_MAGIC	0xda96b2f6
+	int			sock;
+	char			*name;
+	ev_io			listener;
+	struct sockaddr_storage	addr;
 	VTAILQ_ENTRY(listen_sock)	list;
-	ev_io				listener;
-	int				sock;
-	char				*name;
-	struct cfg_cert_file		*cert;
-	struct sslctx_s			*sctx;
-	char				*pspec;
-	struct sockaddr_storage		addr;
 };
 
 VTAILQ_HEAD(listen_sock_head, listen_sock);
 
-static struct listen_sock_head listen_socks;
-static unsigned nlisten_socks;
+struct frontend {
+	unsigned		magic;
+#define FRONTEND_MAGIC	 	0x5b04e577
+	int			match_global_certs;
+	int			sni_nomatch_abort;
+	struct sni_name_s	*sni_names;
+	struct sslctx_s		*ssl_ctxs;
+	char			*pspec;
+	struct listen_sock_head	socks;
+	VTAILQ_ENTRY(frontend)	list;
+};
+
+VTAILQ_HEAD(frontend_head, frontend);
+
+static struct frontend_head frontends;
 
 #define LOG_REOPEN_INTERVAL 60
 static FILE* logf;
@@ -230,7 +241,7 @@ static sni_name *sni_names;
 static sslctx *ssl_ctxs;
 static sslctx *default_ctx;
 
-static void insert_sni_names(sslctx *sc);
+static void insert_sni_names(sslctx *sc, sni_name **sn_tab);
 static int load_cert_ctx(sslctx *so);
 #endif /* OPENSSL_NO_TLSEXT */
 
@@ -582,7 +593,7 @@ handle_shcupd(struct ev_loop *loop, ev_io *w, int revents)
 		r -= sizeof(uint32_t);
 		encdate = *((uint32_t *)&msg[r]);
 		if (!(abs((int)(int32_t)now - ntohl(encdate))
-			< SSL_CTX_get_timeout(default_ctx)))
+			< SSL_CTX_get_timeout(default_ctx->ctx)))
 			continue;
 
 		shctx_sess_add(msg, r, now);
@@ -857,23 +868,12 @@ sni_match(const sni_name *sn, const char *srvname)
 	}
 }
 
-/*
- * Switch the context of the current SSL object to the most appropriate one
- * based on the SNI header
- */
-static int
-sni_switch_ctx(SSL *ssl, int *al, void *data)
+static sslctx *
+sni_lookup(const char *servername, const sni_name *sn_tab)
 {
-	(void)data;
-	(void)al;
-	const char *servername;
 	const sni_name *sn;
 
-	servername = SSL_get_servername(ssl, TLSEXT_NAMETYPE_host_name);
-	if (!servername)
-		return SSL_TLSEXT_ERR_NOACK;
-
-	HASH_FIND_STR(sni_names, servername, sn);
+	HASH_FIND_STR(sn_tab, servername, sn);
 	if (sn == NULL) {
 		char *s;
 		/* attempt another lookup for wildcard matches */
@@ -885,18 +885,59 @@ sni_switch_ctx(SSL *ssl, int *al, void *data)
 
 	if (sn != NULL) {
 		CHECK_OBJ_NOTNULL(sn, SNI_NAME_MAGIC);
-		if (sni_match(sn, servername)) {
-			CHECK_OBJ_NOTNULL(sn->sctx, SSLCTX_MAGIC);
-			SSL_set_SSL_CTX(ssl, sn->sctx->ctx);
-			return SSL_TLSEXT_ERR_OK;
-		}
+		if (sni_match(sn, servername))
+			return (sn->sctx);
 	}
 
+	return (NULL);
+}
+
+/*
+ * Switch the context of the current SSL object to the most appropriate one
+ * based on the SNI header
+ */
+static int
+sni_switch_ctx(SSL *ssl, int *al, void *data)
+{
+	const char *servername;
+	sslctx *sc;
+	const struct frontend *fr = NULL;
+	int lookup_global = 1;
+	int sni_nomatch_abort = CONFIG->SNI_NOMATCH_ABORT;
+
+	(void)al;
+	if (data != NULL)
+		CAST_OBJ_NOTNULL(fr, data, FRONTEND_MAGIC);
+
+	servername = SSL_get_servername(ssl, TLSEXT_NAMETYPE_host_name);
+	if (!servername)
+		return (SSL_TLSEXT_ERR_NOACK);
+
+#define TRY_SNI_MATCH(sn_tab)					\
+	do {							\
+		sc = sni_lookup(servername, (sn_tab));		\
+		if (sc != NULL) {				\
+			CHECK_OBJ_NOTNULL(sc, SSLCTX_MAGIC);	\
+			SSL_set_SSL_CTX(ssl, sc->ctx);		\
+			return (SSL_TLSEXT_ERR_OK);		\
+		}						\
+	} while (0)
+
+	if (fr != NULL) {
+		TRY_SNI_MATCH(fr->sni_names);
+		lookup_global = fr->match_global_certs;
+		if (fr->sni_nomatch_abort != -1)
+			sni_nomatch_abort = fr->sni_nomatch_abort;
+	}
+
+	if (lookup_global)
+		TRY_SNI_MATCH(sni_names);
+
 	/* No matching certs */
-	if (CONFIG->SNI_NOMATCH_ABORT)
-		return SSL_TLSEXT_ERR_ALERT_FATAL;
+	if (sni_nomatch_abort)
+		return (SSL_TLSEXT_ERR_ALERT_FATAL);
 	else
-		return SSL_TLSEXT_ERR_NOACK;
+		return (SSL_TLSEXT_ERR_NOACK);
 }
 #endif /* OPENSSL_NO_TLSEXT */
 
@@ -905,6 +946,9 @@ sctx_free(sslctx *sc, sni_name **sn_tab)
 {
 	sni_name *sn, *sntmp;
 
+	if (sc == NULL)
+		return;
+
 	if (sn_tab != NULL)
 		CHECK_OBJ_NOTNULL(*sn_tab, SNI_NAME_MAGIC);
 	CHECK_OBJ_NOTNULL(sc, SSLCTX_MAGIC);
@@ -923,12 +967,25 @@ sctx_free(sslctx *sc, sni_name **sn_tab)
 }
 
 /* Initialize an SSL context */
-sslctx *
-make_ctx(const struct cfg_cert_file *cf)
+static sslctx *
+make_ctx_fr(const struct cfg_cert_file *cf, const struct frontend *fr,
+    const struct front_arg *fa)
 {
 	SSL_CTX *ctx;
 	sslctx *sc;
 	RSA *rsa;
+	ENC_TYPE etype = CONFIG->ETYPE;
+	char *ciphers = CONFIG->CIPHER_SUITE;
+	int pref_srv_ciphers = CONFIG->PREFER_SERVER_CIPHERS;
+
+	if (fa != NULL) {
+		CHECK_OBJ_NOTNULL(fa, FRONT_ARG_MAGIC);
+		etype = fa->etype;
+		if (fa->ciphers != NULL)
+			ciphers = fa->ciphers;
+		if (fa->prefer_server_ciphers != -1)
+			pref_srv_ciphers = fa->prefer_server_ciphers;
+	}
 
 	long ssloptions = SSL_OP_NO_SSLv2 | SSL_OP_ALL |
 	    SSL_OP_NO_SESSION_RESUMPTION_ON_RENEGOTIATION;
@@ -936,7 +993,7 @@ make_ctx(const struct cfg_cert_file *cf)
 #ifdef SSL_OP_NO_COMPRESSION
 	ssloptions |= SSL_OP_NO_COMPRESSION;
 #endif
-	if (CONFIG->ETYPE == ENC_TLS)
+	if (etype == ENC_TLS)
 		ssloptions |= SSL_OP_NO_SSLv3;
 	ctx = SSL_CTX_new((CONFIG->PMODE == SSL_CLIENT) ?
 	    SSLv23_client_method() : SSLv23_server_method());
@@ -944,13 +1001,13 @@ make_ctx(const struct cfg_cert_file *cf)
 	SSL_CTX_set_options(ctx, ssloptions);
 	SSL_CTX_set_info_callback(ctx, info_callback);
 
-	if (CONFIG->CIPHER_SUITE) {
-		if (SSL_CTX_set_cipher_list(ctx, CONFIG->CIPHER_SUITE) != 1) {
+	if (ciphers != NULL) {
+		if (SSL_CTX_set_cipher_list(ctx, ciphers) != 1) {
 			ERR_print_errors_fp(stderr);
 		}
 	}
 
-	if (CONFIG->PREFER_SERVER_CIPHERS)
+	if (pref_srv_ciphers)
 		SSL_CTX_set_options(ctx, SSL_OP_CIPHER_SERVER_PREFERENCE);
 
 	ALLOC_OBJ(sc, SSLCTX_MAGIC);
@@ -994,6 +1051,10 @@ make_ctx(const struct cfg_cert_file *cf)
 	if (!SSL_CTX_set_tlsext_servername_callback(ctx, sni_switch_ctx)) {
 		ERR("Error setting up SNI support.\n");
 	}
+	CHECK_OBJ_ORNULL(fr, FRONTEND_MAGIC);
+	if (!SSL_CTX_set_tlsext_servername_arg(ctx, fr)) {
+		ERR("Error setting SNI servername arg.\n");
+	}
 #endif /* OPENSSL_NO_TLSEXT */
 
 #ifdef USE_SHARED_CACHE
@@ -1026,7 +1087,7 @@ make_ctx(const struct cfg_cert_file *cf)
 }
 
 static void
-insert_sni_names(sslctx *sc)
+insert_sni_names(sslctx *sc, sni_name **sn_tab)
 {
 	sni_name *sn, *sn2;
 	char *key;
@@ -1037,16 +1098,21 @@ insert_sni_names(sslctx *sc)
 		key = sn->servername;
 		if (sn->is_wildcard)
 			key = sn->servername + 1;
-		HASH_FIND_STR(sni_names, key, sn2);
+		HASH_FIND_STR(*sn_tab, key, sn2);
 		if (sn2 != NULL) {
 			ERR("Warning: SNI name '%s' from '%s' overriden"
 			    " by '%s'\n",
 			    key, sn2->sctx->filename, sn->sctx->filename);
 		}
-		HASH_ADD_KEYPTR(hh, sni_names, key, strlen(key), sn);
+		HASH_ADD_KEYPTR(hh, *sn_tab, key, strlen(key), sn);
 	}
 }
 
+static sslctx *
+make_ctx(const struct cfg_cert_file *cf)
+{
+	return (make_ctx_fr(cf, NULL, NULL));
+}
 #ifndef OPENSSL_NO_TLSEXT
 static int
 load_cert_ctx(sslctx *so)
@@ -1130,57 +1196,9 @@ find_ctx(const char *file)
 void
 init_openssl(void)
 {
-	struct cfg_cert_file *cf, *cftmp;
-	struct listen_sock *ls;
-	sslctx *so;
-
 	SSL_library_init();
 	SSL_load_error_strings();
 
-	AN(CONFIG->CERT_DEFAULT);
-
-	/* The last file listed in config is the "default" cert */
-	default_ctx = make_ctx(CONFIG->CERT_DEFAULT);
-	if (default_ctx == NULL)
-		exit(1);
-
-#ifndef OPENSSL_NO_TLSEXT
-	load_cert_ctx(default_ctx);
-	insert_sni_names(default_ctx);
-
-	// Go through the list of PEMs and make some SSL contexts for
-	// them. We also keep track of the names associated with each
-	// cert so we can do SNI on them later
-	HASH_ITER(hh, CONFIG->CERT_FILES, cf, cftmp) {
-		if (find_ctx(cf->filename) == NULL) {
-			so = make_ctx(cf);
-			if (so == NULL)
-				exit(1);
-			HASH_ADD_KEYPTR(hh, ssl_ctxs, cf->filename,
-			    strlen(cf->filename), so);
-			load_cert_ctx(so);
-			insert_sni_names(so);
-		}
-	}
-
-	VTAILQ_FOREACH(ls, &listen_socks, list) {
-		if (ls->cert) {
-			so = find_ctx(ls->cert->filename);
-			if (so == NULL) {
-				so = make_ctx(ls->cert);
-				AN(so);
-				HASH_ADD_KEYPTR(hh, ssl_ctxs,
-				    ls->cert->filename,
-				    strlen(ls->cert->filename), so);
-				load_cert_ctx(so);
-				insert_sni_names(so);
-			}
-			ls->sctx = so;
-		}
-	}
-
-#undef APPEND_CTX
-#endif /* OPENSSL_NO_TLSEXT */
 
 	if (CONFIG->ENGINE) {
 		ENGINE *e = NULL;
@@ -1203,62 +1221,107 @@ init_openssl(void)
 }
 
 static void
-destroy_listen_sock(struct listen_sock *ls)
+init_certs(void) {
+	struct cfg_cert_file *cf, *cftmp;
+	sslctx *so;
+
+	if (CONFIG->CERT_DEFAULT != NULL) {
+		AN(CONFIG->CERT_DEFAULT);
+		default_ctx = make_ctx(CONFIG->CERT_DEFAULT);
+		if (default_ctx == NULL)
+			exit(1);
+#ifndef OPENSSL_NO_TLSEXT
+		load_cert_ctx(default_ctx);
+		insert_sni_names(default_ctx, &sni_names);
+#endif
+	}
+
+	// Go through the list of PEMs and make some SSL contexts for
+	// them. We also keep track of the names associated with each
+	// cert so we can do SNI on them later
+	HASH_ITER(hh, CONFIG->CERT_FILES, cf, cftmp) {
+		if (find_ctx(cf->filename) == NULL) {
+			so = make_ctx(cf);
+			if (so == NULL)
+				exit(1);
+			HASH_ADD_KEYPTR(hh, ssl_ctxs, cf->filename,
+			    strlen(cf->filename), so);
+#ifndef OPENSSL_NO_TLSEXT
+			load_cert_ctx(so);
+			insert_sni_names(so, &sni_names);
+#endif
+		}
+	}
+}
+
+static void
+destroy_lsock(struct listen_sock *ls)
 {
 	CHECK_OBJ_NOTNULL(ls, LISTEN_SOCK_MAGIC);
 	if (ls->sock > 0)
 		(void) close(ls->sock);
 	free(ls->name);
-	free(ls->pspec);
-	if (ls->cert) {
-		CHECK_OBJ_NOTNULL(ls->cert, CFG_CERT_FILE_MAGIC);
-		ls->cert->ref--;
-		if (ls->cert->ref == 0) {
-			free(ls->cert->filename);
-			FREE_OBJ(ls->cert);
-		}
-	}
 	FREE_OBJ(ls);
 }
 
+static void
+destroy_frontend(struct frontend *fr)
+{
+	struct listen_sock *ls, *lstmp;
+	sslctx *sc, *sctmp;
+
+	CHECK_OBJ_NOTNULL(fr, FRONTEND_MAGIC);
+
+	VTAILQ_FOREACH_SAFE(ls, &fr->socks, list, lstmp) {
+		CHECK_OBJ_NOTNULL(ls, LISTEN_SOCK_MAGIC);
+		VTAILQ_REMOVE(&fr->socks, ls, list);
+		destroy_lsock(ls);
+	}
+
+	HASH_ITER(hh, fr->ssl_ctxs, sc, sctmp) {
+		HASH_DEL(fr->ssl_ctxs, sc);
+		sctx_free(sc, &fr->sni_names);
+	}
+
+	AZ(HASH_COUNT(fr->sni_names));
+	free(fr->pspec);
+	FREE_OBJ(fr);
+}
+
 /* Create the bound socket in the parent process */
 static int
-create_listen_sock(const struct front_arg *fa, struct listen_sock_head *socks)
+frontend_listen(const struct front_arg *fa, struct listen_sock_head *slist)
 {
 	struct addrinfo *ai, hints, *it;
 	struct listen_sock *ls, *lstmp;
 	char buf[INET6_ADDRSTRLEN+20];
 	char abuf[INET6_ADDRSTRLEN];
 	char pbuf[8];
-	int n, r;
-	int count = 0;
-	struct listen_sock_head tmp_list;
+	int r, count = 0;
 
 	CHECK_OBJ_NOTNULL(fa, FRONT_ARG_MAGIC);
-
-	VTAILQ_INIT(&tmp_list);
 	memset(&hints, 0, sizeof hints);
 	hints.ai_family = AF_UNSPEC;
 	hints.ai_socktype = SOCK_STREAM;
 	hints.ai_flags = AI_PASSIVE | AI_ADDRCONFIG;
-	const int gai_err = getaddrinfo(fa->ip, fa->port,
+	r = getaddrinfo(fa->ip, fa->port,
 	    &hints, &ai);
-	if (gai_err != 0) {
+	if (r != 0) {
 		ERR("{getaddrinfo}: %s: [%s]\n", fa->pspec,
-		    gai_strerror(gai_err));
+		    gai_strerror(r));
 		return (-1);
 	}
 
 	for (it = ai; it != NULL; it = it->ai_next) {
 		ALLOC_OBJ(ls, LISTEN_SOCK_MAGIC);
-		VTAILQ_INSERT_TAIL(&tmp_list, ls, list);
+		VTAILQ_INSERT_TAIL(slist, ls, list);
 		count++;
 
 		ls->sock = socket(it->ai_family, SOCK_STREAM, IPPROTO_TCP);
 		if (ls->sock == -1) {
 			ERR("{socket: main}: %s: %s\n", strerror(errno),
 			    fa->pspec);
-			goto creat_ls_err;
+			goto creat_frontend_err;
 		}
 
 		int t = 1;
@@ -1267,7 +1330,7 @@ create_listen_sock(const struct front_arg *fa, struct listen_sock_head *socks)
 		    < 0) {
 			ERR("{setsockopt-reuseaddr}: %s: %s\n", strerror(errno),
 			    fa->pspec);
-			goto creat_ls_err;
+			goto creat_frontend_err;
 		}
 #ifdef SO_REUSEPORT
 		if (setsockopt(ls->sock, SOL_SOCKET, SO_REUSEPORT,
@@ -1275,13 +1338,13 @@ create_listen_sock(const struct front_arg *fa, struct listen_sock_head *socks)
 		    < 0) {
 			ERR("{setsockopt-reuseport}: %s: %s\n", strerror(errno),
 			    fa->pspec);
-			goto creat_ls_err;
+			goto creat_frontend_err;
 		}
 #endif
 		if(setnonblocking(ls->sock) < 0) {
 			ERR("{listen sock: setnonblocking}: %s: %s\n",
 			    strerror(errno), fa->pspec);
-			goto creat_ls_err;
+			goto creat_frontend_err;
 		}
 #ifdef IPV6_V6ONLY
 		t = 1;
@@ -1290,7 +1353,7 @@ create_listen_sock(const struct front_arg *fa, struct listen_sock_head *socks)
 			sizeof (t)) != 0) {
 			ERR("{setsockopt-ipv6only}: %s: %s\n", strerror(errno),
 			    fa->pspec);
-			goto creat_ls_err;
+			goto creat_frontend_err;
 		}
 #endif
 		if (CONFIG->RECV_BUFSIZE > 0) {
@@ -1299,8 +1362,8 @@ create_listen_sock(const struct front_arg *fa, struct listen_sock_head *socks)
 			    sizeof(CONFIG->RECV_BUFSIZE));
 			if (r < 0) {
 				ERR("{setsockopt-rcvbuf}: %s: %s\n",
-				    strerror(errno),fa->pspec);
-				goto creat_ls_err;
+				    strerror(errno), fa->pspec);
+				goto creat_frontend_err;
 			}
 		}
 		if (CONFIG->SEND_BUFSIZE > 0) {
@@ -1310,14 +1373,14 @@ create_listen_sock(const struct front_arg *fa, struct listen_sock_head *socks)
 			if (r < 0) {
 				ERR("{setsockopt-sndbuf}: %s: %s\n",
 				    strerror(errno), fa->pspec);
-				goto creat_ls_err;
+				goto creat_frontend_err;
 			}
 		}
 
 		if (bind(ls->sock, it->ai_addr, it->ai_addrlen)) {
 			ERR("{bind-socket}: %s: %s\n", strerror(errno),
 			    fa->pspec);
-			goto creat_ls_err;
+			goto creat_frontend_err;
 		}
 
 #ifndef NO_DEFER_ACCEPT
@@ -1327,24 +1390,24 @@ create_listen_sock(const struct front_arg *fa, struct listen_sock_head *socks)
 			&timeout, sizeof(int)) < 0) {
 			ERR("{setsockopt-defer_accept}: %s: %s\n",
 			    strerror(errno), fa->pspec);
-			goto creat_ls_err;
+			goto creat_frontend_err;
 		}
 #endif /* TCP_DEFER_ACCEPT */
 #endif
 		if (listen(ls->sock, CONFIG->BACKLOG) != 0) {
 			ERR("{listen-socket}: %s: %s\n", strerror(errno),
 			    fa->pspec);
-			goto creat_ls_err;
+			goto creat_frontend_err;
 		}
 
 		memcpy(&ls->addr, it->ai_addr, it->ai_addrlen);
 
-		n = getnameinfo(it->ai_addr, it->ai_addrlen, abuf,
+		r = getnameinfo(it->ai_addr, it->ai_addrlen, abuf,
 		    sizeof abuf, pbuf, sizeof pbuf,
 		    NI_NUMERICHOST | NI_NUMERICSERV);
-		if (n != 0) {
+		if (r != 0) {
 			ERR("{getnameinfo}: %s\n", fa->pspec);
-			goto creat_ls_err;
+			goto creat_frontend_err;
 		}
 
 		if (it->ai_addr->sa_family == AF_INET6) {
@@ -1354,33 +1417,67 @@ create_listen_sock(const struct front_arg *fa, struct listen_sock_head *socks)
 		}
 		ls->name = strdup(buf);
 		AN(ls->name);
-		if (fa->cert != NULL) {
-			ls->cert = fa->cert;
-			fa->cert->ref++;
-		}
-		ls->pspec = strdup(fa->pspec);
-
 		LOG("{core} Listening on %s\n", ls->name);
 	}
 
-	VTAILQ_FOREACH_SAFE(ls, &tmp_list, list, lstmp) {
-		VTAILQ_REMOVE(&tmp_list, ls, list);
-		VTAILQ_INSERT_TAIL(socks, ls, list);
-	}
-
 	freeaddrinfo(ai);
 	return (count);
 
-creat_ls_err:
+creat_frontend_err:
 	freeaddrinfo(ai);
-	VTAILQ_FOREACH_SAFE(ls, &tmp_list, list, lstmp) {
-		VTAILQ_REMOVE(&tmp_list, ls, list);
-		destroy_listen_sock(ls);
+	VTAILQ_FOREACH_SAFE(ls, slist, list, lstmp) {
+		VTAILQ_REMOVE(slist, ls, list);
+		free(ls->name);
+		if (ls->sock > 0)
+			(void) close(ls->sock);
+		FREE_OBJ(ls);
 	}
 
 	return (-1);
 }
 
+static struct frontend *
+create_frontend(const struct front_arg *fa)
+{
+	struct frontend *fr;
+	sslctx *so;
+	int count = 0;
+	struct frontend_head tmp_list;
+	struct cfg_cert_file *cf, *cftmp;
+
+	CHECK_OBJ_NOTNULL(fa, FRONT_ARG_MAGIC);
+	ALLOC_OBJ(fr, FRONTEND_MAGIC);
+	VTAILQ_INIT(&fr->socks);
+	AN(fr);
+
+	fr->pspec = strdup(fa->pspec);
+	fr->match_global_certs = fa->match_global_certs;
+	fr->sni_nomatch_abort = fa->sni_nomatch_abort;
+
+	VTAILQ_INIT(&tmp_list);
+	count = frontend_listen(fa, &fr->socks);
+	if (count < 0) {
+		destroy_frontend(fr);
+		return (NULL);
+	}
+
+	HASH_ITER(hh, fa->certs, cf, cftmp) {
+		so = make_ctx_fr(cf, fr, fa);
+		if (so == NULL) {
+			destroy_frontend(fr);
+			return (NULL);
+		}
+		HASH_ADD_KEYPTR(hh, fr->ssl_ctxs,
+		    cf->filename, strlen(cf->filename), so);
+#ifndef OPENSSL_NO_TLSEXT
+		load_cert_ctx(so);
+		insert_sni_names(so, &fr->sni_names);
+#endif
+	}
+
+	return (fr);
+}
+
 /* Initiate a clear-text nonblocking connect() to the backend IP on behalf
  * of a newly connected upstream (encrypted) client */
 static int
@@ -2088,6 +2185,7 @@ handle_accept(struct ev_loop *loop, ev_io *w, int revents)
 	(void)loop;
 	struct sockaddr_storage addr;
 	sslctx *so;
+	struct frontend *fr;
 	proxystate *ps;
 	socklen_t sl = sizeof(addr);
 
@@ -2150,7 +2248,12 @@ handle_accept(struct ev_loop *loop, ev_io *w, int revents)
 		return;
 	}
 
-	CAST_OBJ_NOTNULL(so, w->data, SSLCTX_MAGIC);
+	CAST_OBJ_NOTNULL(fr, w->data, FRONTEND_MAGIC);
+	if (fr->ssl_ctxs != NULL)
+		CAST_OBJ_NOTNULL(so, fr->ssl_ctxs, SSLCTX_MAGIC);
+	else
+		CAST_OBJ_NOTNULL(so, default_ctx, SSLCTX_MAGIC);
+
 	SSL *ssl = SSL_new(so->ctx);
 	long mode = SSL_MODE_ENABLE_PARTIAL_WRITE;
 #ifdef SSL_MODE_RELEASE_BUFFERS
@@ -2224,6 +2327,7 @@ handle_accept(struct ev_loop *loop, ev_io *w, int revents)
 static void
 check_ppid(struct ev_loop *loop, ev_timer *w, int revents)
 {
+	struct frontend *fr;
 	struct listen_sock *ls;
 	(void)revents;
 	pid_t ppid = getppid();
@@ -2231,10 +2335,13 @@ check_ppid(struct ev_loop *loop, ev_timer *w, int revents)
 		ERR("{core} Process %d detected parent death, "
 		    "closing listener sockets.\n", core_id);
 		ev_timer_stop(loop, w);
-		VTAILQ_FOREACH(ls, &listen_socks, list) {
-			CHECK_OBJ_NOTNULL(ls, LISTEN_SOCK_MAGIC);
-			ev_io_stop(loop, &ls->listener);
-			close(ls->sock);
+		VTAILQ_FOREACH(fr, &frontends, list) {
+			CHECK_OBJ_NOTNULL(fr, FRONTEND_MAGIC);
+			VTAILQ_FOREACH(ls, &fr->socks, list) {
+				CHECK_OBJ_NOTNULL(ls, LISTEN_SOCK_MAGIC);
+				ev_io_stop(loop, &ls->listener);
+				close(ls->sock);
+			}
 		}
 	}
 }
@@ -2244,6 +2351,7 @@ handle_mgt_rd(struct ev_loop *loop, ev_io *w, int revents)
 {
 	unsigned cg;
 	ssize_t r;
+	struct frontend *fr;
 	struct listen_sock *ls;
 
 	(void) revents;
@@ -2268,10 +2376,13 @@ handle_mgt_rd(struct ev_loop *loop, ev_io *w, int revents)
 		child_state = CHILD_EXITING;
 
 		/* Stop accepting new connections. */
-		VTAILQ_FOREACH(ls, &listen_socks, list) {
-			CHECK_OBJ_NOTNULL(ls, LISTEN_SOCK_MAGIC);
-			ev_io_stop(loop, &ls->listener);
-			close(ls->sock);
+		VTAILQ_FOREACH(fr, &frontends, list) {
+			CHECK_OBJ_NOTNULL(fr, FRONTEND_MAGIC);
+			VTAILQ_FOREACH(ls, &fr->socks, list) {
+				CHECK_OBJ_NOTNULL(ls, LISTEN_SOCK_MAGIC);
+				ev_io_stop(loop, &ls->listener);
+				close(ls->sock);
+			}
 		}
 
 		check_exit_state();
@@ -2287,6 +2398,7 @@ handle_clear_accept(struct ev_loop *loop, ev_io *w, int revents)
 	(void)revents;
 	(void)loop;
 	struct sockaddr_storage addr;
+	struct frontend *fr;
 	sslctx *so;
 	proxystate *ps;
 	socklen_t sl = sizeof(addr);
@@ -2344,7 +2456,11 @@ handle_clear_accept(struct ev_loop *loop, ev_io *w, int revents)
 		return;
 	}
 
-	CAST_OBJ_NOTNULL(so, w->data, SSLCTX_MAGIC);
+	CAST_OBJ_NOTNULL(fr, w->data, FRONTEND_MAGIC);
+	if (fr->ssl_ctxs != NULL)
+		CAST_OBJ_NOTNULL(so, fr->ssl_ctxs, SSLCTX_MAGIC);
+	else
+		CAST_OBJ_NOTNULL(so, default_ctx, SSLCTX_MAGIC);
 	SSL *ssl = SSL_new(so->ctx);
 	long mode = SSL_MODE_ENABLE_PARTIAL_WRITE;
 #ifdef SSL_MODE_RELEASE_BUFFERS
@@ -2410,6 +2526,7 @@ handle_clear_accept(struct ev_loop *loop, ev_io *w, int revents)
 static void
 handle_connections(int mgt_fd)
 {
+	struct frontend *fr;
 	struct listen_sock *ls;
 	struct sigaction sa;
 
@@ -2445,16 +2562,15 @@ handle_connections(int mgt_fd)
 	ev_timer_init(&timer_ppid_check, check_ppid, 1.0, 1.0);
 	ev_timer_start(loop, &timer_ppid_check);
 
-	VTAILQ_FOREACH(ls, &listen_socks, list) {
-		ev_io_init(&ls->listener,
-		    (CONFIG->PMODE == SSL_CLIENT) ?
-		    handle_clear_accept : handle_accept,
-		    ls->sock, EV_READ);
-		if (ls->sctx)
-			ls->listener.data = ls->sctx;
-		else
-			ls->listener.data = default_ctx;
-		ev_io_start(loop, &ls->listener);
+	VTAILQ_FOREACH(fr, &frontends, list) {
+		VTAILQ_FOREACH(ls, &fr->socks, list) {
+			ev_io_init(&ls->listener,
+			    (CONFIG->PMODE == SSL_CLIENT) ?
+			    handle_clear_accept : handle_accept,
+			    ls->sock, EV_READ);
+			ls->listener.data = fr;
+			ev_io_start(loop, &ls->listener);
+		}
 	}
 
 	AZ(setnonblocking(mgt_fd));
@@ -2505,7 +2621,7 @@ init_globals(void)
 	/* backaddr */
 	struct addrinfo hints;
 
-	VTAILQ_INIT(&listen_socks);
+	VTAILQ_INIT(&frontends);
 	VTAILQ_INIT(&child_procs);
 
 	memset(&hints, 0, sizeof hints);
@@ -2823,7 +2939,7 @@ remove_pfh(void)
 struct cfg_tpc_obj;
 
 enum cfg_tpc_type {
-	CFG_LISTEN_ADDR,
+	CFG_FRONTEND,
 	CFG_CERT
 
 	/* ... */
@@ -2865,7 +2981,8 @@ VTAILQ_HEAD(cfg_tpc_obj_head, cfg_tpc_obj);
 
 static struct cfg_tpc_obj *
 make_cfg_obj(enum cfg_tpc_type type, enum cfg_tpc_handling handling,
-    void *priv, cfg_tpc_rollback_f *rollback, cfg_tpc_commit_f *commit)
+    void *priv0, void *priv1, cfg_tpc_rollback_f *rollback,
+    cfg_tpc_commit_f *commit)
 {
 	struct cfg_tpc_obj *o;
 
@@ -2873,7 +2990,8 @@ make_cfg_obj(enum cfg_tpc_type type, enum cfg_tpc_handling handling,
 	AN(o);
 	o->type = type;
 	o->handling = handling;
-	o->p[0] = priv;
+	o->p[0] = priv0;
+	o->p[1] = priv1;
 	o->rollback = rollback;
 	o->commit = commit;
 
@@ -2881,81 +2999,120 @@ make_cfg_obj(enum cfg_tpc_type type, enum cfg_tpc_handling handling,
 }
 
 static void
-ls_rollback(struct cfg_tpc_obj *o)
+frontend_rollback(struct cfg_tpc_obj *o)
 {
-	struct listen_sock *ls;
+	struct frontend *fr;
 
 	if (o->handling == CFG_TPC_NEW) {
-		CAST_OBJ_NOTNULL(ls, o->p[0], LISTEN_SOCK_MAGIC);
-		destroy_listen_sock(ls);
+		CAST_OBJ_NOTNULL(fr, o->p[0], FRONTEND_MAGIC);
+		destroy_frontend(fr);
 	}
 
 	/* KEEP/DROP: ignore */
 }
 
 static void
-ls_commit(struct cfg_tpc_obj *o)
+frontend_commit(struct cfg_tpc_obj *o)
 {
-	struct listen_sock *ls;
+	struct frontend *fr;
 	sslctx *sc;
-	CAST_OBJ_NOTNULL(ls, o->p[0], LISTEN_SOCK_MAGIC);
+	CAST_OBJ_NOTNULL(fr, o->p[0], FRONTEND_MAGIC);
 
 	switch (o->handling) {
 	case CFG_TPC_NEW:
-		VTAILQ_INSERT_TAIL(&listen_socks, ls, list);
+		VTAILQ_INSERT_TAIL(&frontends, fr, list);
 		/* FALL-THROUGH */
 	case CFG_TPC_KEEP:
 		if (o->p[1]) {
 			CAST_OBJ_NOTNULL(sc, o->p[1], SSLCTX_MAGIC);
-			ls->sctx = sc;
+			/* fr->sctx = sc; */
 		}
 		break;
 	case CFG_TPC_DROP:
-		VTAILQ_REMOVE(&listen_socks, ls, list);
-		destroy_listen_sock(ls);
+		VTAILQ_REMOVE(&frontends, fr, list);
+		destroy_frontend(fr);
 		break;
 	}
 }
 
+static void cert_rollback(struct cfg_tpc_obj *o);
+static void cert_commit(struct cfg_tpc_obj *o);
+
+/* Query frontend-specific certificates.  */
+static int
+cert_fr_query(struct frontend *fr, struct front_arg *fa,
+    struct cfg_tpc_obj_head *cfg_objs)
+{
+	struct cfg_cert_file *cf, *cftmp;
+	sslctx *sc, *sctmp;
+	struct cfg_tpc_obj *o;
+
+	HASH_ITER(hh, fr->ssl_ctxs, sc, sctmp) {
+		HASH_FIND_STR(fa->certs, sc->filename, cf);
+		if (cf != NULL && cf->mtim <= sc->mtim) {
+			cf->mark = 1;
+		} else {
+			o = make_cfg_obj(CFG_CERT, CFG_TPC_DROP,
+			    sc, fr, cert_rollback, cert_commit);
+			VTAILQ_INSERT_TAIL(cfg_objs, o, list);
+		}
+	}
+
+	HASH_ITER(hh, fa->certs, cf, cftmp) {
+		if (cf->mark)
+			continue;
+		sc = make_ctx_fr(cf, fr, fa);
+		if (sc == NULL)
+			return (-1);
+		if (load_cert_ctx(sc) != 0) {
+			sctx_free(sc, NULL);
+			return (-1);
+		}
+		o = make_cfg_obj(CFG_CERT, CFG_TPC_NEW,
+		    sc, fr, cert_rollback, cert_commit);
+		VTAILQ_INSERT_TAIL(cfg_objs, o, list);
+	}
+
+	return (0);
+}
+
 /* Query reload of listen sockets.
    Returns -1 on failure.
    Failure: Caller calls .rollback() on the objects added in cfg_objs.
    Success: Caller calls .commit()
 */
 static int
-ls_query(struct front_arg *new_set, struct cfg_tpc_obj_head *cfg_objs)
+frontend_query(struct front_arg *new_set, struct cfg_tpc_obj_head *cfg_objs)
 {
-	struct listen_sock *ls, *lstmp;
+	struct frontend *fr;
 	struct front_arg *fa, *ftmp;
 	struct cfg_tpc_obj *o;
-	struct listen_sock_head lsocks;
-	int n;
 
-	VTAILQ_INIT(&lsocks);
-
-	VTAILQ_FOREACH(ls, &listen_socks, list) {
-		HASH_FIND_STR(new_set, ls->pspec, fa);
+	VTAILQ_FOREACH(fr, &frontends, list) {
+		HASH_FIND_STR(new_set, fr->pspec, fa);
 		if (fa != NULL) {
 			fa->mark = 1;
-			o = make_cfg_obj(CFG_LISTEN_ADDR, CFG_TPC_KEEP, ls,
-			    ls_rollback, ls_commit);
+			o = make_cfg_obj(CFG_FRONTEND, CFG_TPC_KEEP, fr, NULL,
+			    frontend_rollback, frontend_commit);
+			if(cert_fr_query(fr, fa, cfg_objs) < 0) {
+				FREE_OBJ(o);
+				return (-1);
+			}
+
 		} else
-			o = make_cfg_obj(CFG_LISTEN_ADDR, CFG_TPC_DROP, ls,
-			    ls_rollback, ls_commit);
+			o = make_cfg_obj(CFG_FRONTEND, CFG_TPC_DROP, fr, NULL,
+			    frontend_rollback, frontend_commit);
 		VTAILQ_INSERT_TAIL(cfg_objs, o, list);
 	}
 
 	HASH_ITER(hh, new_set, fa, ftmp) {
 		if (!fa->mark) {
-			n = create_listen_sock(fa, &lsocks);
-			if (n <= 0)
+			fr = create_frontend(fa);
+			if (fr == NULL)
 				return (-1);
-			VTAILQ_FOREACH_SAFE(ls, &lsocks, list, lstmp) {
-				VTAILQ_REMOVE(&lsocks, ls, list);
-				o = make_cfg_obj(CFG_LISTEN_ADDR, CFG_TPC_NEW,
-				    ls, ls_rollback, ls_commit);
-				VTAILQ_INSERT_TAIL(cfg_objs, o, list);
-			}
+			o = make_cfg_obj(CFG_FRONTEND, CFG_TPC_NEW,
+			    fr, NULL, frontend_rollback, frontend_commit);
+			VTAILQ_INSERT_TAIL(cfg_objs, o, list);
 		}
 	}
 
@@ -2978,21 +3135,32 @@ static void
 cert_commit(struct cfg_tpc_obj *o)
 {
 	sslctx *sc;
+	sni_name **sn_tab;
+	sslctx **ctxs;
+	struct frontend *fr;
+
+	sn_tab = &sni_names;
+	ctxs = &ssl_ctxs;
 
 	CAST_OBJ_NOTNULL(sc, o->p[0], SSLCTX_MAGIC);
+	if (o->p[1] != NULL) {
+		CAST_OBJ_NOTNULL(fr, o->p[1], FRONTEND_MAGIC);
+		sn_tab = &fr->sni_names;
+		ctxs = &fr->ssl_ctxs;
+	}
 
 	switch (o->handling) {
 	case CFG_TPC_NEW:
-		HASH_ADD_KEYPTR(hh, ssl_ctxs, sc->filename,
+		HASH_ADD_KEYPTR(hh, *ctxs, sc->filename,
 		    strlen(sc->filename), sc);
-		insert_sni_names(sc);
+		insert_sni_names(sc, sn_tab);
 		break;
 	case CFG_TPC_KEEP:
 		WRONG("unreachable");
 		break;
 	case CFG_TPC_DROP:
-		HASH_DEL(ssl_ctxs, sc);
-		sctx_free(sc, &sni_names);
+		HASH_DEL(*ctxs, sc);
+		sctx_free(sc, sn_tab);
 		break;
 	}
 }
@@ -3015,7 +3183,7 @@ dcert_commit(struct cfg_tpc_obj *o)
 	case CFG_TPC_NEW:
 		sctx_free(default_ctx, &sni_names);
 		default_ctx = sc;
-		insert_sni_names(sc);
+		insert_sni_names(sc, &sni_names);
 		break;
 	case CFG_TPC_KEEP:
 		/* FALL-THROUGH */
@@ -3027,14 +3195,14 @@ dcert_commit(struct cfg_tpc_obj *o)
 	}
 }
 
+
 /* Query reload of certificate files */
 static int
 cert_query(hitch_config *cfg, struct cfg_tpc_obj_head *cfg_objs)
 {
 	struct cfg_cert_file *cf, *cftmp;
 	sslctx *sc, *sctmp;
-	struct cfg_tpc_obj *o, *otmp;
-	struct listen_sock *ls;
+	struct cfg_tpc_obj *o;
 
 	/* NB: The ordering here is significant. It is imperative that
 	 * all DROP objects are inserted before any NEW objects, in
@@ -3043,31 +3211,31 @@ cert_query(hitch_config *cfg, struct cfg_tpc_obj_head *cfg_objs)
 		HASH_FIND_STR(cfg->CERT_FILES, sc->filename, cf);
 		if (cf != NULL && cf->mtim <= sc->mtim) {
 			cf->mark = 1;
-			cf->priv = sc;
 		} else {
 			o = make_cfg_obj(CFG_CERT, CFG_TPC_DROP,
-			    sc, cert_rollback, cert_commit);
+			    sc, NULL, cert_rollback, cert_commit);
 			VTAILQ_INSERT_TAIL(cfg_objs, o, list);
 		}
 	}
 
 	/* handle default cert. Default cert has its own
 	 * rollback/commit functions. */
-	AN(cfg->CERT_DEFAULT);
-	cf = cfg->CERT_DEFAULT;
-	CHECK_OBJ_NOTNULL(default_ctx, SSLCTX_MAGIC);
-	if (strcmp(default_ctx->filename, cf->filename) != 0
-	    || cf->mtim > default_ctx->mtim) {
-		sc = make_ctx(cf);
-		if (sc == NULL)
-			return (-1);
-		if (load_cert_ctx(sc) != 0) {
-			sctx_free(sc, NULL);
-			return (-1);
+	if (cfg->CERT_DEFAULT != NULL) {
+		cf = cfg->CERT_DEFAULT;
+		CHECK_OBJ_NOTNULL(default_ctx, SSLCTX_MAGIC);
+		if (strcmp(default_ctx->filename, cf->filename) != 0
+		    || cf->mtim > default_ctx->mtim) {
+			sc = make_ctx(cf);
+			if (sc == NULL)
+				return (-1);
+			if (load_cert_ctx(sc) != 0) {
+				sctx_free(sc, NULL);
+				return (-1);
+			}
+			o = make_cfg_obj(CFG_CERT, CFG_TPC_NEW,
+			    sc, NULL, dcert_rollback, dcert_commit);
+			VTAILQ_INSERT_TAIL(cfg_objs, o, list);
 		}
-		o = make_cfg_obj(CFG_CERT, CFG_TPC_NEW,
-		    sc, dcert_rollback, dcert_commit);
-		VTAILQ_INSERT_TAIL(cfg_objs, o, list);
 	}
 
 	HASH_ITER(hh, cfg->CERT_FILES, cf, cftmp) {
@@ -3080,41 +3248,11 @@ cert_query(hitch_config *cfg, struct cfg_tpc_obj_head *cfg_objs)
 			sctx_free(sc, NULL);
 			return (-1);
 		}
-		cf->priv = sc;
 		o = make_cfg_obj(CFG_CERT, CFG_TPC_NEW,
-		    sc, cert_rollback, cert_commit);
+		    sc, NULL, cert_rollback, cert_commit);
 		VTAILQ_INSERT_TAIL(cfg_objs, o, list);
 	}
 
-	/* This is a little hacky, but we need to also initialize any
-	 * new certs specific to the already queried listen
-	 * endpoints. */
-	VTAILQ_FOREACH(o, cfg_objs, list) {
-		if (o->type != CFG_LISTEN_ADDR)
-			break;
-		if (o->handling == CFG_TPC_DROP)
-			continue;
-		CAST_OBJ_NOTNULL(ls, o->p[0], LISTEN_SOCK_MAGIC);
-		if (ls->cert == NULL)
-			continue;
-		CHECK_OBJ_NOTNULL(ls->cert, CFG_CERT_FILE_MAGIC);
-		HASH_FIND_STR(cfg->CERT_FILES, ls->cert->filename, cf);
-		if (cf != NULL) {
-			CAST_OBJ_NOTNULL(sc, cf->priv, SSLCTX_MAGIC);
-		} else {
-			sc = make_ctx(ls->cert);
-			if (sc == NULL)
-				return (-1);
-			if (load_cert_ctx(sc) != 0) {
-				sctx_free(sc, NULL);
-				return (-1);
-			}
-			otmp = make_cfg_obj(CFG_CERT, CFG_TPC_NEW,
-			    sc, cert_rollback, cert_commit);
-			VTAILQ_INSERT_TAIL(cfg_objs, otmp, list);
-		}
-		o->p[1] = sc;
-	}
 	return (0);
 }
 
@@ -3144,7 +3282,7 @@ reconfigure(int argc, char **argv)
 
 	/* NB: the ordering of the foo_query() calls here is
 	 * significant. */
-	if (ls_query(cfg_new->LISTEN_ARGS, &cfg_objs) < 0
+	if (frontend_query(cfg_new->LISTEN_ARGS, &cfg_objs) < 0
 	    || cert_query(cfg_new, &cfg_objs) < 0) {
 		VTAILQ_FOREACH_SAFE(cto, &cfg_objs, list, cto_tmp) {
 			VTAILQ_REMOVE(&cfg_objs, cto, list);
@@ -3206,7 +3344,7 @@ main(int argc, char **argv)
 
 	// parse command line
 	if (config_parse_cli(argc, argv, CONFIG, &rv) != 0) {
-		fprintf(stderr, "%s", config_error_get());
+		fprintf(stderr, "%s\n", config_error_get());
 		return (rv);
 	}
 
@@ -3215,6 +3353,7 @@ main(int argc, char **argv)
 		    " certificates\n");
 		init_globals();
 		init_openssl();
+		init_certs();
 		fprintf(stderr, "%s configuration looks ok.\n",
 		    basename(argv[0]));
 		return (0);
@@ -3248,16 +3387,21 @@ main(int argc, char **argv)
 	openssl_check_version();
 
 	init_signals();
-
 	init_globals();
+	init_openssl();
 
 	HASH_ITER(hh, CONFIG->LISTEN_ARGS, fa, ftmp) {
-		int c = create_listen_sock(fa, &listen_socks);
-		if (c <= 0)
+		struct frontend *fr = create_frontend(fa);
+		if (fr == NULL)
 			exit(1);
-		nlisten_socks += c;
+		VTAILQ_INSERT_TAIL(&frontends, fr, list);
 	}
 
+	/* load certificates, pass to handle_connections */
+	LOGL("{core} Loading certificate pem files (%d)\n",
+	    HASH_COUNT(CONFIG->CERT_FILES) + 1); /* XXX: TODO */
+	init_certs();
+
 #ifdef USE_SHARED_CACHE
 	if (CONFIG->SHCUPD_PORT) {
 		/* create socket to send(children) and
@@ -3266,11 +3410,6 @@ main(int argc, char **argv)
 	}
 #endif /* USE_SHARED_CACHE */
 
-	/* load certificates, pass to handle_connections */
-	LOGL("{core} Loading certificate pem files (%d)\n",
-	    HASH_COUNT(CONFIG->CERT_FILES) + 1);
-	init_openssl();
-
 	if (CONFIG->CHROOT && CONFIG->CHROOT[0])
 		change_root();
 
